use std::sync::mpsc::channel;

use termion::event::{Event, Key, MouseEvent};
use termion::input::TermRead;

use crate::coordinates::{Coordinates, Position, Size};
use crate::fail::{HResult, HError};
use crate::window::{send_event, Events};

use std::io::{BufWriter, Write, stdin};


pub trait Widget {
    fn get_widget(&self) -> Box<dyn Widget> {
        Box::new(crate::textview::TextView::new_blank())
    }
    fn get_coordinates(&self) -> &Coordinates;
    fn set_coordinates(&mut self, coordinates: &Coordinates);
    fn render_header(&self) -> String;
    fn render_footer(&self) -> String { "".into() }
    fn refresh(&mut self);
    fn get_drawlist(&self) -> String;
    fn after_draw(&self) -> HResult<()> { Ok(()) }


    fn on_event(&mut self, event: Event) -> HResult<()> {
        match event {
            Event::Key(Key::Char('q')) => panic!("It's your fault!"),
            Event::Key(key) => self.on_key(key),
            Event::Mouse(button) => self.on_mouse(button),
            Event::Unsupported(wtf) => self.on_wtf(wtf),
        }
    }

    fn on_key(&mut self, key: Key) -> HResult<()> {
        match key {
            _ => self.bad(Event::Key(key)),
        }
        Ok(())
    }

    fn on_mouse(&mut self, event: MouseEvent) -> HResult<()> {
        match event {
            _ => self.bad(Event::Mouse(event)),
        }
        Ok(())
    }

    fn on_wtf(&mut self, event: Vec<u8>) -> HResult<()> {
        match event {
            _ => self.bad(Event::Unsupported(event)),
        }
        Ok(())
    }

    fn show_status(&self, status: &str) {
        crate::window::show_status(status);
    }

    fn minibuffer(&self, query: &str) -> HResult<String> {
        crate::window::minibuffer(query)
    }

    fn bad(&mut self, event: Event) {
        self.show_status(&format!("Stop the nasty stuff!! {:?} does nothing!", event));
    }

    fn get_header_drawlist(&mut self) -> String {
        format!(
            "{}{}{:xsize$}{}{}",
            crate::term::goto_xy(1, 1),
            crate::term::header_color(),
            " ",
            crate::term::goto_xy(1, 1),
            self.render_header(),
            xsize = self.get_coordinates().xsize() as usize
        )
    }

    fn get_footer_drawlist(&mut self) -> String {
        let xsize = self.get_coordinates().xsize();
        let ypos = crate::term::ysize();
        format!(
            "{}{}{:xsize$}{}{}",
            crate::term::goto_xy(1, ypos),
            crate::term::header_color(),
            " ",
            crate::term::goto_xy(1, ypos),
            self.render_footer(),
            xsize = xsize as usize)
    }

    fn get_clearlist(&self) -> String {
        let (xpos, ypos) = self.get_coordinates().u16position();
        let (xsize, ysize) = self.get_coordinates().u16size();

        (ypos..ysize + 2)
            .map(|line| {
                format!(
                    "{}{}{:xsize$}",
                    crate::term::reset(),
                    crate::term::goto_xy(xpos, line),
                    " ",
                    xsize = xsize as usize
                )
            })
            .collect()
    }

    fn get_redraw_empty_list(&self, lines: usize) -> String {
        let (xpos, ypos) = self.get_coordinates().u16position();
        let (xsize, ysize) = self.get_coordinates().u16size();

        let start_y = lines + ypos as usize;
        (start_y..(ysize + 2) as usize)
            .map(|i| {
                format!(
                    "{}{:xsize$}",
                    crate::term::goto_xy(xpos, i as u16),
                    " ",
                    xsize = xsize as usize
                )
            })
            .collect()
    }

    fn draw(&self) -> HResult<()> {
        let drawlist = self.get_drawlist();
        let mut bufout = BufWriter::new(std::io::stdout());

        write!(bufout, "{}", drawlist)?;
        bufout.flush()?;
        Ok(())
    }

    fn popup(&mut self) -> HResult<()> {
        self.run_widget();
        send_event(Events::ExclusiveEvent(None))?;
        Ok(())
    }

    fn run_widget(&mut self) -> HResult<()> {
        let (tx_event, rx_event) = channel();
        send_event(Events::ExclusiveEvent(Some(tx_event)))?;

        self.clear()?;
        self.refresh();
        self.draw()?;

        for event in rx_event.iter() {
            match event {
                Events::InputEvent(input) => {
                    if let Err(HError::PopupFinnished) = self.on_event(input) {
                        return Err(HError::PopupFinnished)
                    }
                }
                Events::WidgetReady => {
                    self.refresh();
                }
                _ => {}
            }
            self.draw();
            self.after_draw();
        }
        Ok(())
    }

    fn clear(&self) -> HResult<()> {
        let clearlist = self.get_clearlist();
        write!(std::io::stdout(), "{}", clearlist)?;
        std::io::stdout().flush()?;
        Ok(())
    }

    fn animate_slide_up(&mut self) {
        let coords = self.get_coordinates().clone();
        let xpos = coords.position().x();
        let ypos = coords.position().y();
        let xsize = coords.xsize();
        let ysize = coords.ysize();
        let clear = self.get_clearlist();
        let pause = std::time::Duration::from_millis(5);
        let mut bufout = BufWriter::new(std::io::stdout());

        for i in (0..10).rev() {
            let coords = Coordinates { size: Size((xsize,ysize-i)),
                                       position: Position
                                           ((xpos,
                                             ypos+i))
            };
            self.set_coordinates(&coords);
            let buffer = self.get_drawlist();
            write!(bufout, "{}{}",
                   clear, buffer).unwrap();
            bufout.flush().ok();

            std::thread::sleep(pause);
        }
    }
}
